/******************************************************************************
*
* (c) 2004 by BECK IPC GmbH
*
* BECK IPC GmbH
* Garbenheimerstr. 36-38
* D-35578 Wetzlar
*
* Phone: (49)-6441-905-0
* Fax:   (49)-6441-905-245
*
*******************************************************************************
*
* Disclaimer: This program is an example and should be used as such.
*             If you wish to use this program or parts of it in your application,
*             you must validate the code yourself. BECK IPC GmbH can not be held
*             responsible for the correct functioning or coding of this example.
*
* Module:     taskexp1.c
* Function:   This program demonstrates the usage of the RTOS-Interface of the
*             IPC@CHIP.
*             At the start of the program we demonstrate the usage of the set
*             time/date function, by setting time and date of the system.
*             After that we create two tasks, which print a message with the
*             current time/date. The printing function is protected by a
*             resource semaphore.
*             After about 20 seconds the program exits, the created tasks will be
*             removed.
*
*******************************************************************************
*
* $Header: c:\proj2000\IPC@CHIP\Examples\rtos\TASKEXP1\TASKEXP1.C, 5, 31.03.2004 11:16:02, Andre Pribil$
*
******************************************************************************/

/******************************************************************************
* Includes
******************************************************************************/
#include <stdio.h>
#include <clib.h>

/******************************************************************************
* Constants
******************************************************************************/
#define TASK_STACKSIZE  1024   // WORDS => 2048 Bytes, was 1024
#define TASK1_DEBUG   // debug switch

/******************************************************************************
* Global variables and prototypes
******************************************************************************/

void huge task1(void);
void huge task2(void);

unsigned int task1_stack[TASK_STACKSIZE];
int task1ID;

TaskDefBlock  task1defblock =
{
  task1,
  {'T','A','S','1'},             // a name: 4 chars
  &task1_stack[TASK_STACKSIZE],  // top of stack
  TASK_STACKSIZE*sizeof(int),   // size of stack
  0,                             // attributes, not supported
  26,                            // priority 3(high) ... 127(low)
  0,                             // time slice (if any), not supported
  0,0,0,0                        // mailbox depth, not supported
};

unsigned int task2_stack[TASK_STACKSIZE];
int task2ID;

TaskDefBlock  task2defblock =
{
  task2,
  {'T','A','S','2'},             // a name: 4 chars
  &task2_stack[TASK_STACKSIZE],  // top of stack
  TASK_STACKSIZE*sizeof(int),   // size of stack
  0,                             // attributes, not supported
  27,                            // priority 3(high) ... 127(low)
  0,                             // time slice (if any), not supported
  0,0,0,0                        // mailbox depth, not supported
};

// Init a variable for setting the system time
TimeDate_Structure td =
{
   0,    /* seconds  (0-59)   */
   0,    /* minutes  (0-59)   */
   16,   /* hours (0-23)      */
   20,   /* day      (1-31)   */
   1,    /* month (1-12)      */
   0,    /* year     (0-99)   */
   3,    /* day of week (Mon=1 to Sun=7)     */
   20    /* century */
};

TimeDate_Structure tdget;

int   print_semaphoreID; /*semaphoreID*/
long  semaphore_timout = 0L;
char  print_sem[5] = {'P','S','E','1',0};
int   task1run;
int   task2run;



/******************************************************************************
* print_date_time()
*
* Prints date and time to STDOUT
******************************************************************************/
void print_date_time(TimeDate_Structure *td)
{
  printf("  Time: %02d:%02d:%02d",td->hr,td->min,td->sec);
  printf(" ,Date: %02d.%02d.%04d\r\n",td->dy,td->mn,td->yr+td->dcen*100);
}



/******************************************************************************
* print_task_message()
*
* This function will be called periodic by the two tasks.
* We protect this function with a resource semaphore, because we don't want to
* to interrupt this function.
******************************************************************************/
void print_task_message(char * taskmsg)
{
  int result=0;
  long time =0L;

  //reserve the resource, wait forever
  result =  RTX_Reserve_Sem(print_semaphoreID,&time);

  if(result==0)
  {
    printf("\r\n%s ",taskmsg);
    RTX_Get_TimeDate(&tdget);
    print_date_time(&tdget);

  }
  RTX_Release_Sem(print_semaphoreID);//release the resource
}



/******************************************************************************
* task1()
*
* This function will be executed in task 1
******************************************************************************/
void huge task1(void)
{

   int result;

   task1run=1;

   while(task1run)
   {
       RTX_Sleep_Time(500);    /*sleep*/
       print_task_message("TASK1");
   }

    /*
      You should not delete a task, which is waiting, or is in about
      a wait, for a resource or counting semaphore.
      It should be safe to delete the task by itself after the
      while loop.
    */

   printf("\r\nKill task1");
   result = RTX_Kill_Task(RTX_Get_TaskID());
   if(result!=0)
   {
      printf(" failed, %d  %d\r\n",result,task1ID);
   }
}



/******************************************************************************
* task2()
*
* This function will be executed in task 2
******************************************************************************/
void huge task2(void)
{
   int result;

   task2run=1;
   while(task2run)
   {
       RTX_Sleep_Time(800); /*sleep */
       print_task_message("TASK2");
   }

   /*
      You should not delete a task, which is waiting, or is in about
      a wait, for a resource or counting semaphore.
      It should be safe to delete the task by itself after the
      while loop.
   */



   printf("\r\nKill task2");
   result = RTX_Kill_Task(RTX_Get_TaskID());
   if(result!=0)
   {
      printf(" failed, %d\r\n",result);
   }
}



/******************************************************************************
* main()
******************************************************************************/
int main(void)
{
   int result;
   int loopcnt=0;
   static int firsttime=1;
   //switch stdio to our own program
   BIOS_Set_Focus( FOCUS_APPLICATION );


   printf("\r\nRTOS API example1\r\n");
   RTX_Set_TimeDate(&td);
   RTX_Get_TimeDate(&tdget);


   //create a resource semaphore for function print_task_msg
   result = RTX_Create_Sem(&print_semaphoreID,print_sem,-1);

   if(result!=0)
   {
     printf("\r\nCreating the semaphore failed %d, exit program\r\n",result);
     return 0;
   }

   //create/restart and start task1 and task2
   if(firsttime==1)
      result = RTX_Create_Task(&task1ID , &task1defblock);
   else
      result = RTX_Restart_Task(task1ID);
   if(result!=0)
   {
     printf("\r\nCreating/restart TASK_1 failed %d, exit program\r\n",result);
     //delete print_semaphore
     RTX_Delete_Sem(print_semaphoreID);
     return 0;
   }

   //create/restart and start task1 and task2
   if(firsttime==1)
     result = RTX_Create_Task(&task2ID , &task2defblock);
   else
      result = RTX_Restart_Task(task2ID);


   firsttime=0;


   if(result!=0)
   {
     printf("\r\nCreating/restart TASK_2 failed %d, exit program\r\n",result);
     //delete task1
     RTX_Delete_Task(task1ID);
     //delete print_semaphore
     RTX_Delete_Sem(print_semaphoreID);
     return 0;
   }

   loopcnt=0;
   while(loopcnt<20)
   {

       RTX_Sleep_Time(1000);
       loopcnt++;
   }

   /*
      Tell task1 and task2 to exit their while loop
   */
   task1run = task2run = 0;
   RTX_Sleep_Time(3000);
   printf("\r\nDeleting the semaphore ");
   result = RTX_Delete_Sem(print_semaphoreID);
   if(result!=0)
   {
      printf(" failed, %d\r\n",result);
   }

   printf("\r\nExit program\r\n\r\n");

   //switch stdio to command shell
   BIOS_Set_Focus( FOCUS_BOTH );


   return 0;
}



// End of file
