/******************************************************************************
*
* (c) 2003-2004 by BECK IPC GmbH
*
* BECK IPC GmbH
* Garbenheimerstr. 36-38
* D-35578 Wetzlar
*
* Phone: (49)-6441-905-0
* Fax:   (49)-6441-905-245
*
*******************************************************************************
*
* Disclaimer: This program is an example and should be used as such.
*             If you wish to use this program or parts of it in your application,
*             you must validate the code yourself. BECK IPC GmbH can not be held
*             responsible for the correct functioning or coding of this example.
*
* Module:     sntp.c
* Function:   This program is a simple SNTP client that queries the actual date
*             and time from a NTP server and sets the system clock if the user
*             whishes.
*             The IP address of the NTP server is expected as command line
*             parameter. To get a usage message call the program without any
*             parameters.
* Notes:      To get a better grasp of what is done in the function getTime()
*             you should gallop through RFC 1769.
*
*******************************************************************************
*
* $Header: Sntp.c, 3, 19.08.2004 10:26:25, Jan Schatz$
*
******************************************************************************/



/******************************************************************************
* Includes
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <clib.h>
#include <string.h>
#include <dos.h>
#include "bool.h"



/******************************************************************************
* Constants
******************************************************************************/
#define VERBOSE
#define NTP_PORT             123          // Well known SNTP port
#define TIMEZONE_CORRECTION  5            // NTP doesn't deliver UTC but UTC - 5
#define OFFSET_JAN_1_1970    2208988800L  // Difference between Jan 1st 1900
                                          // 0:00:00 (NTP's base) and Jan 1st
                                          // 1970 0:00:00 (Unix Timestamp's base)
                                          // in seconds



/******************************************************************************
* Type definitions
******************************************************************************/
typedef struct
{
  unsigned char li_vn_mode;
  unsigned char stratum;
  unsigned char poll;
  unsigned char precision;
  unsigned long rootDelay;
  unsigned long rootDispersion;
  unsigned long referenceIdentifier;
  unsigned long reference_Timestamp[2];
  unsigned long originateTimestamp[2];
  unsigned long receiveTimestamp[2];
  unsigned long transmitTimestamp[2];
  unsigned long authenticator[3];
} TSntpHeader;



/******************************************************************************
* headerGetLi()
*
* Gets the LI field from the given TSntpHeader-structure.
*
* Returns the value of the LI field or 0xFF if no valid TSntpHeader-structure
* is given.
******************************************************************************/
unsigned char headerGetLi(
                           TSntpHeader *sntpHeader )  // TSntpHeader-structure to
                                                      // get field from
{
  if( sntpHeader )
    return ( sntpHeader->li_vn_mode & 0xC0 ) >> 6;
  else return 0xFF;
}



/******************************************************************************
* headerSetLi()
*
* Sets the LI field in the given TSntpHeader-structure.
******************************************************************************/
void headerSetLi(
                  unsigned char li,          // Value of the LI field
                  TSntpHeader *sntpHeader )  // TSntpHeader-structure to set field in
{
  if( sntpHeader &&
      li <= 3 )
    sntpHeader->li_vn_mode |= ( li & 0x03 ) << 6;
}



/******************************************************************************
* headerSetVn()
*
* Sets the VN field in the given TSntpHeader-structure.
******************************************************************************/
void headerSetVn(
                  unsigned char vn,          // Value of the VN field
                  TSntpHeader *sntpHeader )  // TSntpHeader-structure to set field in
{
  if( sntpHeader &&
      vn <= 7 )
    sntpHeader->li_vn_mode |= ( vn & 0x07 ) << 3;
}



/******************************************************************************
* headerSetMode()
*
* Sets the Mode field in the given TSntpHeader-structure.
******************************************************************************/
void headerSetMode(
                    unsigned char mode,        // Value of the Mode field
                    TSntpHeader *sntpHeader )  // TSntpHeader-structure to set field in
{
  if( sntpHeader &&
      mode <= 7 )
    sntpHeader->li_vn_mode |= ( mode & 0x07 );
}




/******************************************************************************
* getTime()
******************************************************************************/
BOOL getTime(
              struct date *date,
              struct time *time,     // Destination for date and time (DOS format)
              char *remoteIp,        // IP address of the NTP server
              signed int timezone )  // Addend for timezone (e.g. +1 for MET)
{
  static TSntpHeader sntpHeader;  // SNTP header for request/ reply
  struct sockaddr_in address;     // Address structure
  int socket,                     // Socket for connection to server
      error;                      // Result of TCP/IP-API calls
  unsigned long unixTimestamp;    // time in Unix format


  // Check addend for timezone
  if( abs( timezone ) > 12 )
  {
    #ifdef VERBOSE
    printf( "\r\nError: The addend for the timezone is out of range!" );
    #endif
    return FALSE;
  }


  // Initialize header according tp RFC 1769 5
  memset( &sntpHeader, 0, sizeof( TSntpHeader ) );
  headerSetLi( 0, &sntpHeader );  // No warning
  headerSetVn( 3, &sntpHeader );  // NTP Version
  headerSetMode( 3, &sntpHeader );  // Client, unicast


  // Open socket
  if( ( socket = opensocket( SOCK_DGRAM, &error ) ) == API_ERROR )
  {
    #ifdef VERBOSE
    printf( "\r\nError: Could not open socket! (%i)", error );
    #endif
    return FALSE;
  }


  // Prepare and send request
  address.sin_family = AF_INET;
  address.sin_port = htons( NTP_PORT );
  inet_addr( remoteIp, &address.sin_addr.s_addr );
  if( sendto( socket, (char *)&sntpHeader, sizeof( TSntpHeader ), MSG_BLOCKING, &address, &error ) == API_ERROR )
  {
    #ifdef VERBOSE
    printf( "\r\nError: Could not send request! (%i)", error );
    #endif
    closesocket( socket, &error );
    return FALSE;
  }


  // Receive reply
  memset( &sntpHeader, 0, sizeof( TSntpHeader ) );   // Clear SNTP-header
  memset( &address, 0, sizeof( struct sockaddr ) );  // Clear address
  address.sin_family = AF_INET;
  if( recvfrom( socket, (char *)&sntpHeader, sizeof( TSntpHeader ), MSG_TIMEOUT, 30000L, &address, &error ) == API_ERROR )
  {
    #ifdef VERBOSE
    printf( "\r\nError: Could not receive reply! (%i)", error );
    #endif
    closesocket( socket, &error );
    return FALSE;
  }
  closesocket( socket, &error );


  // Check reply
  if( headerGetLi( &sntpHeader ) == 3 ||
      sntpHeader.stratum  < 1 || sntpHeader.stratum > 15 ||
      sntpHeader.transmitTimestamp[0] == 0 ||
      sntpHeader.transmitTimestamp[1] == 0 )
    printf( "\r\nWarning: Clock of NTP server is not synchronized!" );
  

  // Convert time to Unix format
  unixTimestamp = ( sntpHeader.transmitTimestamp[0] & 0xFF000000L ) >> 24;
  unixTimestamp |= ( sntpHeader.transmitTimestamp[0] & 0x00FF0000L ) >> 8;
  unixTimestamp |= ( sntpHeader.transmitTimestamp[0] & 0x0000FF00L ) << 8;
  unixTimestamp |= ( sntpHeader.transmitTimestamp[0] & 0x000000FFL ) << 24;
                   // Correct byte order and discard decimal places
  unixTimestamp -= OFFSET_JAN_1_1970;  // NTP delivers the number of seconds since
                                       // January 1st 1900. Unix format is number
                                       // of seconds since January 1st 1970.

  // Correct time according to timezone
  unixTimestamp += ( timezone + TIMEZONE_CORRECTION ) * 3600;

  // Convert time to DOS format
  unixtodos( unixTimestamp, date, time );


  return TRUE;
}




/******************************************************************************
* main()
******************************************************************************/
#pragma argsused
int main(int argc, char **argv)
{
  char *month[] = { "---", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
                    "Aug", "Sep", "Oct", "Nov", "Dec" };
  struct date date;
  struct time time;
  int i = 0,
      timezone = 0;
  BOOL setClock = FALSE;

	unsigned long ta;

  // Set STDIO focus to application
  BIOS_Set_Focus( FOCUS_APPLICATION );

  // Say hello
  printf( "\r\n--------------------------------------------------"
          "\r\nIPC@Chip SNTP-Demo"
          "\r\n--------------------------------------------------"
          "\r\n" );

   RTX_Get_System_Ticks(&ta);
   ta=ta/1000;  // elapsed time in seconds
	printf ("\r\nCurrent Time in ticks: %lu", ta);
   printf( "\r\nDate: %s %u. %u", month[date.da_mon], date.da_day, date.da_year );
   printf( "\r\nTime: %u:%02u:%02u", time.ti_hour, time.ti_min, time.ti_sec);

  // Check number of command line parameters
  if( argc < 2 )
  {
    printf( "\r\nUsage: SNTP.EXE [-S] [TIMEZONE] REMOTEIP"
            "\r\n -S:        System clock will be set with this option."
            "\r\n TIMEZONE:  Specify the addend for your timezone here (e.g. +1 for MET)."
            "\r\n            Defaults to 0 (UTC)."
            "\r\n REMOTEIP:  IP address of NTP server"
            "\r\n" );
  }
  else
  {
    // Parse command line parameters
    for( i = 1; i < argc - 1; ++i )
    {
      if( timezone == 0 )
        timezone = atoi( argv[i] );

      if( stricmp( argv[i], "-s" ) == 0 )
        setClock = TRUE;
    }


    // Get time
    if( !getTime( &date, &time, argv[argc-1], timezone ) )
    {
      printf( "\r\nError: Could not query current date and time!" );
    }
    else
    {
      printf( "\r\nTimezone: UTC%+i", timezone );
      printf( "\r\nDate: %s %u. %u", month[date.da_mon], date.da_day, date.da_year );
      printf( "\r\nTime: %u:%02u:%02u", time.ti_hour, time.ti_min, time.ti_sec);

      // Set system clock
      if( setClock )
      {
        printf( "\r\nSetting system clock..." );
        setdate( &date );
        settime( &time );

          RTX_Get_System_Ticks(&ta);
 		  ta=ta/1000;  // elapsed time in seconds
			printf ("\r\nCurrent Time in ticks: %lu", ta);

      }
    }
  }


  // Say bye
  printf( "\r\nGood Bye."
          "\r\n" );

  // Set STDIO focus to both
  BIOS_Set_Focus( FOCUS_BOTH );


  return 0;
}



// End of file

